/*******************************************************************************
* Freescale Semiconductor Inc.
* (c) Copyright 2011 Freescale Semiconductor, Inc.
* ALL RIGHTS RESERVED.
********************************************************************************
Services performed by FREESCALE in this matter are performed AS IS and without
any warranty. CUSTOMER retains the final decision relative to the total design
and functionality of the end product. FREESCALE neither guarantees nor will be
held liable by CUSTOMER for the success of this project.
FREESCALE DISCLAIMS ALL WARRANTIES, EXPRESSED, IMPLIED OR STATUTORY INCLUDING,
BUT NOT LIMITED TO, IMPLIED WARRANTY OF MERCHANTABILITY OR FITNESS FOR
A PARTICULAR PURPOSE ON ANY HARDWARE, SOFTWARE ORE ADVISE SUPPLIED 
TO THE PROJECT BY FREESCALE, AND OR NAY PRODUCT RESULTING FROM FREESCALE 
SERVICES. IN NO EVENT SHALL FREESCALE BE LIABLE FOR INCIDENTAL OR CONSEQUENTIAL 
DAMAGES ARISING OUT OF THIS AGREEMENT.
CUSTOMER agrees to hold FREESCALE harmless against any and all claims demands 
or actions by anyone on account of any damage, or injury, whether commercial,
contractual, or tortuous, rising directly or indirectly as a result 
of the advise or assistance supplied CUSTOMER in connection with product, 
services or goods supplied under this Agreement.
********************************************************************************
* File:             main.c
* Owner:            b05111
* Version:          0.0
* Date:             Aug-11-2014
* Classification:   General Business Information
* Brief:            
********************************************************************************
* Detailed Description: 
* Example show simple flash programming routine. During runtime it changes 
* content of field of constants 'test' (thus located in internal flash). 
* Also it shows how to relocate code into RAM a data into FLASH (used linker
* command file is MPC5643L_my_sections.lcf and MPC5643L_DEBUG_my_sections.lcf).
*
* Note: For complex tasks use SSD driver (Freescale site for particular device, 
* Software&Tools/Run-Time Software/Middleware-Device Drivers
*
* ------------------------------------------------------------------------------
* Test HW:         xPC564xLKIT, PPC5643L Cut3 silicon
* Target :         internal_FLASH, RAM
* Fsys:            120 MHz PLL0
* Debugger:        Lauterbach Trace32
*                  PeMicro USB-ML-PPCNEXUS
* Terminal:        19200-8-no parity-1 stop bit-no flow control via LINFlex0
* EVB connection:  default 
*
********************************************************************************
Revision History:
0.0       Aug-11-2014     b05111  Initial Version               
*******************************************************************************/
#include <stdio.h>
#include "MPC5643L.h"
#include "uart.h"
#include "IntcInterrupts.h"

/*******************************************************************************
* External objects
*******************************************************************************/

/*******************************************************************************
* Global variables
*******************************************************************************/


#pragma push // save the current state
#pragma section const_type ".my_flash"
/* align this to a 0-modulo-8 address (aligned to 8 bytes) */
static const vuint32_t test[] __attribute__ ((aligned (8))) = 
{
    0xAAAAAAAA,
    0xBBBBBBBB,
    0xCCCCCCCC,
    0xDDDDDDDD
};
#pragma pop
// Note: 
// at default constant smaller then 8 bytes are placed into .sdata and .sdata2
// bigger data are placed into .rodata (that is flash memory)

const vuint32_t *p_test = &test[0];
vuint32_t test_read;

static vuint32_t data_to_be_written[] = 
{
  	0xFEEDFACE,
    0xBABADEDA,
    0xCAFEBEEF,
    0xDEDABABA
};

static vuint32_t size_of_test = (sizeof(test)); // how many bytes is needed 


/*******************************************************************************
* Constants and macros
*******************************************************************************/

/* FCCU keys */
#define FCCU_NCFK_KEY 0xAB3498FE
#define FCCU_CFK_KEY  0x618B7A50

/* PCR numbers */
#define PD4_pin 52
#define PD5_pin 53  
#define PD6_pin 54
#define PB6_pin 22

#define LED1_pin   PD4_pin
#define LED2_pin   PD5_pin
#define LED3_pin   PD6_pin
#define CLKOUT_pin PB6_pin


/*******************************************************************************
* Local types
*******************************************************************************/

/*******************************************************************************
* Local function prototypes
*******************************************************************************/
static void HW_init(void);
static void SysClk_init(void);
static void ModesAndClks_init(void);
static void PeriClkGen_init(void);
static void DisableWatchdog(void);
static void GPIO_init(void);
static void ClearNCF(void);
static void ClearCF(void);
static void ClearFails(void);
static void PIT3_init(void);
static void PIT3_isr(void);
static void Terminal_ECHO(void);

// load to RAM if FLASH target chosen
#ifdef ROM_VERSION
    #pragma push
    #pragma section code_type ".my_ram" ".my_ram" code_mode=far_abs
#endif
    void Program_FLASH(void);
#ifdef ROM_VERSION
    #pragma pop
#endif

/*******************************************************************************
Function Name : HW_init
Engineer      : b08110
Date          : Feb-18-2010
Parameters    : 
Modifies      : 
Returns       : 
Notes         : initialization of the hw for the purposes of this example
Issues        : 
*******************************************************************************/
static void HW_init(void)
{
    DisableWatchdog();
    ClearFails();
    SysClk_init();    
    GPIO_init();    
}

/*******************************************************************************
Function Name : SysClk_init
Engineer      : b08110
Date          : Feb-25-2010
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void SysClk_init(void)
{
    ModesAndClks_init();
    PeriClkGen_init();
}

/*******************************************************************************
Function Name : ModesAndClks_init
Engineer      : b08110
Date          : Aug-26-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - DRUN->RUN1->RUN0->RUN1
                - enable CLKOUT
Issues        : 
*******************************************************************************/
static void ModesAndClks_init(void) 
{
    int32_t cnt = 0;
    
    /* Enable DRUN, RUN0, RUN1 SAFE, RESET modes */
    ME.MER.R = 0x0000003D;
    
    /* Mode Transition to enter RUN1 mode: */
        
    // RUN1 cfg: 16MHzIRCON,OSC0ON,PLL0OFF,PLL1OFF,syclk=16MIRC
    ME.RUN[1].R = 0x001F0030;
    // Enter RUN1 Mode & Key
    ME.MCTL.R = 0x50005AF0;
    // Enter RUN1 Mode & Inverted Key
    ME.MCTL.R = 0x5000A50F;
    // Wait for mode entry to complete
    while(0 == ME.GS.B.S_XOSC) {};
    // Wait for mode transition to complete
    while(1 == ME.GS.B.S_MTRANS) {}
    // Check RUN1 mode has been entered
    while(5 != ME.GS.B.S_CURRENT_MODE) {};
    
    
    /* Initialize PLL before turning it on:
       fsys = fcrystal*ndiv/idf/odf
       fvco must be from 256MHz to 512MHz
       we want fsys = 120MHz. fvco = fsys*odf = 120MHz * 4 = 480MHz
       fsys = 40*72/6/4 = 120MHz */

    CGM.AC3SC.R = 0x01000000; // Select Xosc as PLL0 source clock
    CGM.AC4SC.R = 0x01000000; // Select Xosc as PLL1 source clock

#if 0
    /* 80 MHz */
    CGM.FMPLL[0].CR.B.IDF = 0x7; // FMPLL0 IDF=5 --> divide by 8
    CGM.FMPLL[0].CR.B.ODF = 0x1; // FMPLL0 ODF=1 --> divide by 4
    CGM.FMPLL[0].CR.B.NDIV = 64; // FMPLL0 NDIV=72 --> divide by 64
    
    /* 120 MHz */
    CGM.FMPLL[1].CR.B.IDF = 0x5; // FMPLL0 IDF=5 --> divide by 8
    CGM.FMPLL[1].CR.B.ODF = 0x1; // FMPLL0 ODF=1 --> divide by 4
    CGM.FMPLL[1].CR.B.NDIV = 72; // FMPLL0 NDIV=72 --> divide by 64
#else
    /* 120 MHz */
    CGM.FMPLL[0].CR.B.IDF = 0x5; // FMPLL0 IDF=5 --> divide by 6
    CGM.FMPLL[0].CR.B.ODF = 0x1; // FMPLL0 ODF=1 --> divide by 4
    CGM.FMPLL[0].CR.B.NDIV = 72; // FMPLL0 NDIV=72 --> divide by 72
    
    /* 80 MHz */
    CGM.FMPLL[1].CR.B.IDF = 0x7; // FMPLL0 IDF=5 --> divide by 8
    CGM.FMPLL[1].CR.B.ODF = 0x1; // FMPLL0 ODF=1 --> divide by 4
    CGM.FMPLL[1].CR.B.NDIV = 64; // FMPLL0 NDIV=72 --> divide by 64
#endif

    CGM.FMPLL[0].CR.B.EN_PLL_SW = 1; // enable progressive clock switching
    CGM.FMPLL[1].CR.B.EN_PLL_SW = 1; // enable progressive clock switching
    
    ME.RUNPC[0].R = 0x000000FE; // enable peripherals run in all modes
    ME.LPPC[0].R = 0x00000000;  // disable peripherals run in LP modes

#if 0     
    /* Mode Transition to enter RUN0 mode: */
    
    // RUN0 cfg: 16MHzIRCON,OSC0ON,syclk=OSC
    ME.RUN[0].R = 0x001F0032;       
    // Enter RUN0 Mode & Key
    ME.MCTL.R = 0x40005AF0;         
    // Enter RUN0 Mode & Inverted Key
    ME.MCTL.R = 0x4000A50F;         
    // Wait for mode transition to complete 
    while (ME.IS.B.I_MTC != 1) {}     
    // Clear Transition flag  
    ME.IS.R = 0x00000001;           
#endif
    
    /* Mode Transition to enter RUN0 mode: */
    
    // RUN0 cfg: 16MHzIRCON,OSC0ON,PLL0ON,PLL1ON,syclk=16M IRC
    ME.RUN[0].R = 0x001F00F0;
    // Enter RUN0 Mode & Key
    ME.MCTL.R = 0x40005AF0;
    // Enter RUN0 Mode & Inverted Key
    ME.MCTL.R = 0x4000A50F;
    // Wait for mode transition to complete
    while (1 == ME.GS.B.S_MTRANS) {}
    // Check RUN0 mode has been entered
    while(4 != ME.GS.B.S_CURRENT_MODE) {};
    
    /* Mode Transition to enter RUN1 mode: */
    
    // RUN1 cfg: 16MHzIRCON,OSC0ON,PLL0ON,PLL1ON,syclk=PLL0
    ME.RUN[1].R = 0x001F00F4;
    // Enter RUN1 Mode & Key
    ME.MCTL.R = 0x50005AF0;
    // Enter RUN1 Mode & Inverted Key
    ME.MCTL.R = 0x5000A50F;
    // Wait for mode transition to complete
    while(1 == ME.GS.B.S_MTRANS) {}
    // Check RUN1 mode has been entered
    while(5 != ME.GS.B.S_CURRENT_MODE) {};
    
    /* enable CLKOUT on PB6 */
    SIU.PCR[CLKOUT_pin].R = 0x0600; // ALT1 - PCR[22] - PA = 0b01
    
    /* set CLKOUT divider of 8 */
    CGM.OCDS_SC.R = 0x32000000; // div by 8, system FMPLL, cut2
    CGM.OCEN.B.EN = 1; 			// enable CLKOUT signal
}

/*******************************************************************************
Function Name : PeriClkGen_init
Engineer      : b08110
Date          : Apr-14-2010
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - Enable auxiliary clock 0-3. from PLL0
Issues        : 
*******************************************************************************/
static void PeriClkGen_init(void) 
{
    CGM.AC0SC.R = 0x04000000;  /* MPC56xxP: Select PLL0 for aux clk 0  */
    CGM.AC0DC.R = 0x80800000;  /* MPC56xxP: Enable aux clk 0 div by 1 */
    CGM.AC1SC.R = 0x04000000;  /* MPC56xxP: Select PLL0 for aux clk 1  */
    CGM.AC1DC.R = 0x80000000;  /* MPC56xxP: Enable aux clk 1 div by 1 */
    CGM.AC2SC.R = 0x04000000;  /* MPC56xxP: Select PLL0 for aux clk 2  */
    CGM.AC2DC.R = 0x80000000;  /* MPC56xxP: Enable aux clk 2 div by 1 */
}

/*******************************************************************************
Function Name : DisableWatchdog
Engineer      : b08110
Date          : Jun-23-2010
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - disable SWT functionality
Issues        : 
*******************************************************************************/
static void DisableWatchdog(void) 
{
    SWT.SR.R = 0x0000c520;     /* Write keys to clear soft lock bit */
    SWT.SR.R = 0x0000d928; 
    SWT.CR.R = 0x8000010A;     /* Clear watchdog enable (WEN) */
    
    /* e200 Core Watchdog Timer */
    asm
    {
        li  r3, 0
        mtspr   340, r3
    }
}

/*******************************************************************************
Function Name : GPIO_init
Engineer      : b08110
Date          : Feb-08-2010
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void GPIO_init(void)
{    	 	
 	SIU.GPDO[LED1_pin].R = 1;
    SIU.GPDO[LED2_pin].R = 1;
    SIU.GPDO[LED3_pin].R = 1;
    
    SIU.PCR[LED1_pin].B.PA = 0x00; 
    SIU.PCR[LED2_pin].B.PA = 0x00;
    SIU.PCR[LED3_pin].B.PA = 0x00;
    
    SIU.PCR[LED1_pin].B.OBE = 0x01;
    SIU.PCR[LED2_pin].B.OBE = 0x01;
    SIU.PCR[LED3_pin].B.OBE = 0x01;    
    
}

/*******************************************************************************
Function Name : ClearNCF
Engineer      : b08110
Date          : Apr-08-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void ClearNCF(void)
{
    uint32_t i,b[4];
    for(i=0;i<4;i++)
    {
        FCCU.NCFK.R = FCCU_NCFK_KEY;
        FCCU.NCF_S[i].R = 0xFFFFFFFF;
        while(FCCU.CTRL.B.OPS != 0x3)
        {
        
        };              /* wait for the completion of the operation */
        b[i]=FCCU.NCF_S[i].R;
    }
}

/*******************************************************************************
Function Name : ClearCF
Engineer      : b08110
Date          : Apr-08-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void ClearCF(void)
{
    uint32_t i,a[4];
    for(i=0;i<4;i++)
    {
        FCCU.CFK.R = FCCU_CFK_KEY;
   
        FCCU.CF_S[i].R = 0xFFFFFFFF;
  
        while(FCCU.CTRL.B.OPS != 0x3)
        {
            
        };      /* wait for the completion of the operation */

        a[i]=FCCU.CF_S[i].R;
    }
}

/*******************************************************************************
Function Name : ClearFails
Engineer      : b08110
Date          : Oct-26-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : 
Issues        : 
*******************************************************************************/
static void ClearFails(void)
{
    uint16_t reset_register;
    
    if(RGM.FES.B.F_FCCU_SAFE /*|| RGM.FES.B.F_FCCU_HARD*/)
    {
        reset_register = RGM.FES.R;
        ME.IMTS.R = 0x00000001;
        /* ClearCF(); */
        ClearNCF();
        
        RGM.FES.R = 0xFFFF;
        RGM.DES.R = 0xFFFF;
        
        /* re-enter DRUN */
        ME.MCTL.R = 0x30005AF0; /* Enter DRUN Mode & Key */        
        ME.MCTL.R = 0x3000A50F; /* Enter DRUN Mode & Inverted Key */
    }
}


/*******************************************************************************
Function Name : PIT3_init
Engineer      : b08110
Date          : Mar-12-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - 1 second periodic interrup using the PIT3 module
Issues        : - expecting 120 MHz fsys for the PIT timer
*******************************************************************************/
static void PIT3_init(void)
{
	/* 30: MDIS = 0 to enable clock for PITs. */
    /* 31: FRZ = 1 for Timers stopped in debug mode */
    PIT.PITMCR.R = 0x00000001;    
    PIT.LDVAL3.R = 120000000 - 1;
    	
    /* clear the TIF flag */
    PIT.TFLG3.R = 0x00000001;
    	
    /* 30: TIE = 1 for interrupt request enabled */
    /* 31: TEN = 1 for timer active */
    PIT.TCTRL3.R = 0x00000003;
}

/*******************************************************************************
Function Name : PIT3_isr
Engineer      : b08110
Date          : Mar-12-2011
Parameters    : 
Modifies      : 
Returns       : 
Notes         : - 1 second periodic interrup using the PIT3 module
Issues        : 
*******************************************************************************/
static void PIT3_isr(void)
{
    /* toggle LED */
    SIU.GPDO[LED3_pin].R ^= 1;

    PIT.TFLG3.R = 0x00000001;    
}


/*******************************************************************************
Function Name : Terminal_ECHO
Engineer      : b05111
Date          : Apr-19-2012
Parameters    : 
Modifies      : 
Returns       : 
Notes         :                          
Issues        : ReadUARTPoll must be defined in uart.c
*******************************************************************************/

static void Terminal_ECHO(void)
{
    vuint32_t result;
    char ch; 
    
    /* if character is present, read it and re-sent it to terminal */
    result = ReadUARTPoll(&ch);
    if (result==kUARTNoError) 
    {
        switch(ch)
        {
            // enter
            case '\n':
            case '\r':
                printf("\r\n");
                break;
                    
            // other symbols
            default:
                printf("%c",ch); 
                fflush(stdout);
        }           
    } 

}


/*******************************************************************************
Function Name : Program_FLASH
Engineer      : b05111
Date          : 
Parameters    : 
Modifies      : 
Returns       : 
Notes         :                          
Issues        : - RAM target preproccesor must NOT define ROM_VERSION macro
                - Internal_FLASH target preproccesor must define ROM_VERSION
                  (Read While Write allowed only for different partitions so
                  relocating of this function to RAM is required for flash 
                  target) 
                - section my_flash is defined in linker file with range
                  0x000C0000-0x000FFFFF thus it fits to block H1
*******************************************************************************/
#ifdef ROM_VERSION
#pragma push
#pragma section code_type ".my_ram" ".my_ram" code_mode=far_abs
#endif
void Program_FLASH(void)
{
    #define FLASH_HLR_PASSWORD      0xB2B22222  

    /* enable high address space */	
    CFLASH.HBL.R = FLASH_HLR_PASSWORD; // unlock register
    CFLASH.HBL.R = 0x00000000;         // unlock all high blocks

    /* step1. erase H1 (0xC0000-0xFFFFF) */
    CFLASH.MCR.B.ERS = 1;
    CFLASH.HBS.R = 0x00000002;               // select H1 block
    *((unsigned int*) p_test) = 0xFFFFFFFF;  // interlock write 
    CFLASH.MCR.B.EHV = 1;                    // start the erase operation
    while(CFLASH.MCR.B.DONE == 0){};         // wait for DONE
    CFLASH.MCR.B.EHV = 0;
    CFLASH.MCR.B.ERS = 0;

    /* step2. program data */
    CFLASH.MCR.B.PGM = 1;              
    *((unsigned int*) p_test++) = data_to_be_written[0]; // first write
    *((unsigned int*) p_test++) = data_to_be_written[1]; // additional write
    *((unsigned int*) p_test++) = data_to_be_written[2]; // additional write
    *((unsigned int*) p_test) = data_to_be_written[3];   // additional write
    CFLASH.MCR.B.EHV = 1;                                //start the erase operation
    while(CFLASH.MCR.B.DONE == 0){}; //wait for DONE
    CFLASH.MCR.B.EHV = 0;
    CFLASH.MCR.B.PGM = 0;
}

#ifdef ROM_VERSION
#pragma pop
#endif


/*******************************************************************************
* Global functions
*******************************************************************************/

int32_t main(void) 
{
    vint32_t counter = 0;
    vint32_t i = 0;

    HW_init(); 
    
    /* Install PIT3_isr into the interrupt vector table */
    INTC_InstallINTCInterruptHandler(PIT3_isr, 127, 1);
    
    /* lower current INTC priority to start INTC interrupts */
    INTC.CPR_PRC0.R = 0;
    
    /* 1 sec periodical interrupt */
    PIT3_init();  

    /* Read content of constant variable 'test' */ 
    
    printf("Initial value of 'test' field is: \r\n");
    printf("0x%x\r\n", test[0]);
    printf("0x%x\r\n", test[1]);
    printf("0x%x\r\n", test[2]);
    printf("0x%x\r\n", test[3]);
   
    /* Program 'test' with new content */
    Program_FLASH();
    
    /* Read programmed data */
    printf("After re-flashing value of 'test' field is: \r\n");
    printf("0x%x\r\n", test[0]);
    printf("0x%x\r\n", test[1]);
    printf("0x%x\r\n", test[2]);
    printf("0x%x\r\n", test[3]);
   
    /* Check programmed data */
    
    for(i=0; i<size_of_test/4; i++)
    {
        if (test[i] != data_to_be_written[i])
        {
             printf("Programming failed\r\n");
             while (1)
             {
                asm (nop);   
             }
        } 
    }
    
    printf("Programming succesful\r\n");
    printf("Have a nice day\r\n");
    
    /* Loop forever */
    while (1) 
    {
        /* delay */
        while(counter < 500000)
        {
            counter++;
            Terminal_ECHO();

        }  
        counter = 0;
        
        /* toggle LED */
        SIU.GPDO[LED1_pin].R ^= 1;
        SIU.GPDO[LED2_pin].R ^= 1;
           
    }
    
}

